#include <string.h>
#include "common.h" // must be on top for INTERNAL_IO definition

#if !defined(INTERNAL_IO)
  #include <stdio.h>
  #include <stdlib.h>
#endif

#if defined(__BORLANDC__)
  #pragma exit ShutDownProgram
#endif

// GLOBAL VARIABLES
BOOL VerboseMode=FALSE;

static PExitFunction ExitFuncs[NUM_EXITPROCS];
static int NumExitFuncs=0;

const char ERROR_IN_MODULE[]="Error in "__FILE__": ";

/**********************************/

void RunItOnExit(PExitFunction p){
  if(NumExitFuncs==NUM_EXITPROCS){
    PrintStringNLF(ERROR_IN_MODULE);
    PrintString("ExitFuncs overflow in \"RunItOnExit\".");
    FinishProgram(ecProgramError); }
  else
    ExitFuncs[NumExitFuncs++]=p;
};

void ShutDownProgram(void){
  if(NumExitFuncs){
    const char *msg[NUM_EXITPROCS];
    int flag=0;
    int i=NumExitFuncs;
    for(--NumExitFuncs; NumExitFuncs>=0; NumExitFuncs--)
      if((msg[NumExitFuncs]=ExitFuncs[NumExitFuncs]())!=NULL) flag=1;
    if(flag){
      PrintString("[PROGRAM SHUTDOWN LOG]");
      for(--i; i>=0; i--) PrintString(msg[i]);
    };
    NumExitFuncs=0;
  };
}

void ExitProgram(int code, const char *s, const char*, ...){
  static int Inside=0;
  if(Inside){
    // 㭪        訡
    PrintStringNLF(ERROR_IN_MODULE);
    PrintString("recursive shutdown (\"ExitProgram\").\n"
                "SHUTDOWN NOT COMPLETE - ABNORMAL PROGRAM TERMINATION");
    FinishProgram(char(ecTotalFault));
  }else{
    const char **sp=&s;
    int flag=0;
    Inside=1;
    const char *msg[NUM_EXITPROCS];
    int i=NumExitFuncs;
    for(--NumExitFuncs; NumExitFuncs>=0; NumExitFuncs--)
      if((msg[NumExitFuncs]=ExitFuncs[NumExitFuncs]())!=NULL) flag=1;
    PrintString("[EXECUTION ERROR]");
    while(*sp!=NULL) PrintStringNLF(*sp++);
    if(flag){
      PrintString("[PROGRAM TERMINATION LOG]");
      for(--i; i>=0; i--) PrintString(msg[i]);
    };
    FinishProgram(code);
  };
}

char* GetPath(const char *src, char *dest){
  strcpy(dest, src);
  char* end=strrchr(dest, PATH_DIVIDER);
  if(end!=NULL) *end=0; else *dest=0;
  return dest;
}

char* AddExt(const char* s, const char* ext, char* out){
  strcpy(out, s);
  char* p=strchr(out, '.');
  if(p==NULL) strcat(strcat(out, "."), ext);
  return out;
}

void* CritAlloc(unsigned s){
  void* ret=new char[s];
  if(ret==NULL)
    ExitProgram(ecUnrecOutOfMemory,
      "Program terminated by \"CritAlloc\": not enough memory.");
  return ret;
}

int TryAlloc(int n, char** p1, unsigned s...){
  PChar adrptr=PChar(&p1), szptr=PChar(&s);
  int cnt=0; // allocated blocks counter
  while(n--){
    if(((*PPChar(*PPChar(adrptr)))=new char[*Punsigned(szptr)])==NULL){
    //                  㬥
    //         㬥 (㪠⥫  PChar)
    //  㥬 㪠⥫ (PChar)
      for(int i=cnt-1; i>=0; i--){ //   ।!
        adrptr-=sizeof(PChar)+sizeof(unsigned);
        delete *PPChar(*PPChar(adrptr));
      };
      return cnt;
    };
    cnt++;
    adrptr+=sizeof(PChar)+sizeof(unsigned);
    szptr +=sizeof(PChar)+sizeof(unsigned);
  };
  return cnt;
}

void PrintChar(char c){
  #if defined(INTERNAL_IO)
    asm mov  ah, 2
    asm mov  dl, c
    asm cmp  dl, 0Ah
    asm jne  JustPrint
    asm mov  dl, 0Dh
    asm int  21h
    asm mov  dl, c
    JustPrint:
    asm int  21h
  #else
    putchar(c);
  #endif
}

void PrintStringNLF(const char *s){
  if(s!=NULL)
    while(*s) PrintChar(*s++);
}

void PrintString(const char *s){
  if(s!=NULL){
    while(*s) PrintChar(*s++);
    if(*--s!='\n') PrintChar('\n');
  }
}

void FinishProgram(char code){
  NumExitFuncs=0;
  #if defined(INTERNAL_IO)
    asm mov  ah, 04Ch
    asm mov  al, code
    asm int  21h
  #else
    exit(code);
  #endif
}

PCChar TObject::ClassName() const {
  return "TObject";
}

#if !defined(FASTRND_TABLE)
  UWORD RandomNumber=0xCDEF;
#else
  UBYTE
    Random_Seed=0,
    Random_Table[256]={ 0x63, 0x58, 0x22, 0x8B, 0x3C, 0x6F, 0x7A, 0x45, 0xFA,
      0xF9, 0x79, 0x25, 0xB5, 0x50, 0xA6, 0x71, 0x28, 0x2B, 0xF5, 0x6A, 0x1B,
      0xD9, 0x4B, 0x5C, 0xF8, 0x48, 0x13, 0x52, 0x76, 0x6C, 0xD1, 0xCA, 0xBA,
      0xE9, 0x12, 0x29, 0x70, 0x42, 0x00, 0x7D, 0x73, 0xD4, 0xF3, 0xFD, 0x16,
      0x2D, 0xC3, 0x90, 0x9B, 0x3E, 0x8C, 0x88, 0xD7, 0x27, 0x3D, 0xFF, 0xDD,
      0x34, 0x09, 0xCB, 0xE2, 0x53, 0xCD, 0x32, 0xCE, 0xC0, 0xFB, 0x2C, 0xEE,
      0x80, 0xC4, 0x46, 0x75, 0xE5, 0x21, 0x4C, 0xED, 0x41, 0x38, 0x43, 0x5F,
      0x4D, 0xD8, 0x98, 0x84, 0x82, 0xEB, 0x57, 0xDC, 0x1C, 0xA7, 0xCC, 0x7F,
      0x18, 0x6D, 0xB6, 0xC1, 0xC6, 0x5A, 0x83, 0xF7, 0xE3, 0x11, 0x78, 0x20,
      0xAE, 0x1E, 0x01, 0x91, 0x4F, 0xF6, 0xE8, 0x40, 0x9F, 0x97, 0xCF, 0x03,
      0x8F, 0x51, 0xDF, 0xAF, 0xB0, 0x26, 0x69, 0x30, 0xEC, 0x5D, 0x59, 0x4E,
      0x62, 0x44, 0xC8, 0xEF, 0x14, 0x92, 0xF1, 0x0F, 0x7B, 0x35, 0xD2, 0x8D,
      0xA0, 0xB7, 0x23, 0x0C, 0x3B, 0x60, 0x72, 0xB3, 0x61, 0xDE, 0x93, 0x81,
      0x89, 0xEA, 0xD0, 0xDA, 0x2F, 0xAA, 0x95, 0xA5, 0xBE, 0xBC, 0xB9, 0x2A,
      0x2E, 0x74, 0xBF, 0x9E, 0xAB, 0x3A, 0xE0, 0x87, 0x9C, 0x07, 0x55, 0xFE,
      0x8E, 0x1A, 0xD5, 0x15, 0x67, 0x7C, 0xB1, 0x8A, 0xAC, 0x37, 0x85, 0xD3,
      0x05, 0x08, 0xD6, 0xE6, 0xE4, 0xF0, 0xB4, 0xA3, 0xF4, 0x65, 0x5E, 0x17,
      0xA8, 0xC7, 0x96, 0x06, 0xE1, 0x24, 0x54, 0x04, 0xC9, 0xC2, 0x77, 0xA1,
      0x1F, 0xA9, 0x0E, 0xBB, 0xC5, 0xFC, 0xF2, 0xB8, 0xAD, 0x0A, 0xE7, 0x39,
      0x49, 0x0D, 0xA4, 0x5B, 0x33, 0x66, 0x31, 0x02, 0x7E, 0x3F, 0x86, 0x19,
      0x6E, 0x56, 0xBD, 0x47, 0xDB, 0xA2, 0x4A, 0x6B, 0x36, 0x10, 0x9A, 0x94,
      0x9D, 0xB2, 0x99, 0x68, 0x0B, 0x1D, 0x64 };
#endif
