#if !defined(COMMON_H)
#define COMMON_H

// uses DOS routines from asm code if defined
// comment out to use standart C I/O functions and exit procedure
#if defined(__BORLANDC__) && defined(__MSDOS__)
  #define INTERNAL_IO
#endif


// '\\' for DOS/Windowz, '/' for UNIX
#define PATH_DIVIDER '\\'

#define MAX_ALLOC_SIZE 65520lu

// Defined:   use random predefined table (256 numbers)
// UNdefined: use simple & fast calculation (65536 numbers)
//#define FASTRND_TABLE

// any file with NULL definition - may be stdio.h
#if defined(__BORLANDC__)
  #include <_null.h>
#else
  #include <stdio.h>
#endif

//  ᮧ ⠭ ᨭ  ⨯ 㪠⥫  뫮  :
#define ClassDefs(name)\
  typedef name* P##name;        \
  typedef const name* PC##name; \
  typedef name& R##name;        \
  typedef const name& RC##name;

/********************** Standart program exit codes *************************/
const int
  ecNormalExit       =0,
  ecBadData          =80,
  ecUnrecFileError   =90,
  ecVideoBIOSError   =100,
  ecIncompartMode    =101,
  ecUnrecOutOfMemory =110, // unable to continue due to out of memory
  // 让 ᡮ  㭪樮஢ ணࠬ, 易  訡  ணࠬ-
  // ஢  ࠧ襭 :
  ecProgramError     =120,
  ecFileOpenBug      =121, // 㬥⭮ ⨥ 䠩
  ecFileNotOpen      =122,
  ecReadOutOfFile    =123,
  ecOutOfBounds      =124,
  ecTotalFault       =128; // huge error

/*****************************************************************************
*+++++++++++++++++++++++++++++++ DATA TYPES +++++++++++++++++++++++++++++++++*
*****************************************************************************/
enum TAllocType { matNoMemory=-1, matStatic=0, matAuto=1 };
enum BOOL       { FALSE=0, TRUE=1, False=0, True=1 };
enum            { NO=0, YES=1};

/************************* STANDART TYPES ALIASES ***************************/
typedef char           Char;
typedef unsigned char  UBYTE;
typedef signed   char  BYTE;
typedef unsigned short UWORD;
typedef signed   short WORD;
typedef unsigned long  UDWORD;
typedef signed   long  DWORD;
typedef unsigned long  ULONG;
typedef unsigned int   UINT;

#if defined(__BORLANDC__)
  #if ((sizeof(BYTE)!=1) || (sizeof(WORD)!=2) || (sizeof(DWORD)!=4))
    #error Please check types definitions in common.h !!!
  #endif
#endif

typedef void*       Pvoid;  // "void& is not a valid type"
typedef const void* PCvoid;
ClassDefs(char);
ClassDefs(Char);
ClassDefs(PChar);
ClassDefs(int);
ClassDefs(unsigned);
ClassDefs(long);
ClassDefs(UBYTE);
ClassDefs(BYTE);
ClassDefs(UWORD);
ClassDefs(WORD);
ClassDefs(UDWORD);
ClassDefs(DWORD);
ClassDefs(ULONG);
ClassDefs(UINT);

ClassDefs(TAllocType);
ClassDefs(BOOL);

// _bios_keybrd(), bioskey() codes
const unsigned
  BK_LEFT   =19200, BK_RIGHT  =19712,
  BK_UP     =18432, BK_DOWN   =20480,
  BK_SPACE  =14624,
  BK_ENTER  =7181,
  BK_BS     =3592,
  BK_TAB    =3849,  BK_SH_TAB =0x0F00,
  BK_INS    =20992,
  BK_DEL    =21248,
  BK_HOME   =18176, BK_PGUP   =18688,
  BK_PGDN   =20736, BK_END    =20224,
  BK_ESC    =283,
  BK_F1     =15104, BK_F2     =15360,
  BK_F3     =15616, BK_F4     =15872,
  BK_F5     =16128, BK_F6     =16384,
  BK_F7     =16640, BK_F8     =16896,
  BK_F9     =17152, BK_F10    =17408;
// Byte codes:
const char DEL=127, ENTER=13, ESC=27;

// Useful inline-functions:
inline char UpCaseLat(char x){ return ((x>='a') && (x<='z'))? (x-' '):x; }
template <class T> inline void Swap(T& a, T& b){ T tmp=a; a=b; b=tmp; }
template <class T> inline T Abs(T x){ return (x<T(0))?(-x):x; }
template <class T> inline T Max(T a, T b){ return (a>b)?a:b; }
template <class T> inline T Min(T a, T b){ return (a>b)?b:a; }
template <class T> inline T Sqr(T x){ return x*x; }
template <class T> inline T Sign(T x){ return (x>=T(0))?T(1):T(-1); }
template <class T> inline T Limit(T x, T lim){ return (x<=lim)?x:lim; }
// faster than random() !
#if defined(FASTRND_TABLE)
  #define FASTRND_MAX 255u
  extern UBYTE Random_Seed, Random_Table[256];
  inline UBYTE FRandom(){ return Random_Table[Random_Seed++]; }
  inline UWORD FRandom(UWORD lim){ return (FRandom()*lim+128)/255; }
#else
  #define FASTRND_MAX 65535u
  extern UWORD RandomNumber;
  inline UWORD FRandom(){ return RandomNumber=RandomNumber*25183+13849; }
  inline UWORD FRandom(UWORD lim){
    return unsigned(UDWORD((FRandom())*UDWORD(lim)+32767lu)/65535lu); }
#endif

// GLOBAL VARIABLES
extern BOOL VerboseMode;

//  㭪権  襭 ணࠬ
const int NUM_EXITPROCS=16;

typedef const char* ExitFunction(void);
typedef ExitFunction* PExitFunction;

#if defined(__MSDOS__)
  typedef void far IntFunctionStub(void);
#else
  typedef void IntFunctionStub(void); 
#endif
typedef IntFunctionStub* PIntFunctionStub;

void RunItOnExit(PExitFunction p);
void ShutDownProgram(void);
void ExitProgram(int code, const char *s, const char*s2=NULL, ...);

char* GetPath(const char *src, char *dest);
char* AddExt(const char* s, const char* ext, char* out);

// ।     new
//  訡 蠥 ணࠬ
void* CritAlloc(unsigned s);
// । ந쭮 ⢮   (p १ s),  訡
// ᢮ ,  㤠 ।;
// 頥 ⢮ ,  㤠 ।
// (᫨   ࠢ 襭,  ॠ쭮   । )
int TryAlloc(int n, char** p1, unsigned s...);

// FOR INTERNAL USE ONLY !!
void PrintChar(char c);
void PrintStringNLF(const char *s);
void PrintString(const char *s); // feeds line only if s does not do
void FinishProgram(char code);

// STANDARD CLASS PROTOTYPES

class TObject{
  public:
    virtual PCChar ClassName() const;
};
ClassDefs(TObject);

#endif
