// Copyright (C) Sergey Galin, 1999-2002

// Can be built in: 
//
//   GCC / Win32
//   Visual C++
//   Borland C++ / MS-DOS

#define VERSIONNUM "1.4"
//#define TRACE

// History:
//
// See CHANGELOG.txt for information.

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#if defined(LINUX)
  #include <sys/types.h>
  #include <dirent.h>   
  #include <sys/stat.h>
  #include <fcntl.h>   
  #include <unistd.h>
  #include "library/common.h"
  #include "library/stdcon.h"
  #define PATH_CHAR '/'
  #define O_BINARY 0
  #define O_TEXT 0
  #define memicmp memcmp  
#else
  #include <dos.h>
  #include <io.h>
  #include <sys\stat.h>
  #include <fcntl.h>
  #include "..\library\common.h"
  #include "..\library\stdcon.h"  
  #define PATH_CHAR '\\'  
#endif  


#if defined(__MSDOS__)
  #define SIM_VERSION "Version: "VERSIONNUM"/MS-DOS"
  #include <dir.h>
  const unsigned attr=FA_RDONLY|FA_HIDDEN|FA_SYSTEM|FA_DIREC|FA_ARCH;
#else

  #if defined(LINUX)
    #define SIM_VERSION "Version: "VERSIONNUM"/Linux"
    #define MAXPATH 65535
  #else
    #if defined(__GNUC__)
      #define SIM_VERSION "Version: "VERSIONNUM"/Win32/GCC"
    #elif defined(_MSC_VER)
      #define SIM_VERSION "Version: "VERSIONNUM"/Win32/MSC"
    #else
      #define SIM_VERSION "Version: "VERSIONNUM"/Win32"
    #endif

    #include <direct.h>
    #define findfirst    _findfirst
    #define findnext     _findnext
    #define ffblk        _finddata_t
    #define getcwd       _getcwd
    #define setdisk      _chdrive
    #define getdisk      _getdrive
    #define ff_attrib    attrib
    #define ff_name      name
    #define ff_fsize     size 
    #define FA_DIREC     _A_SUBDIR
    const unsigned attr=_A_RDONLY|_A_HIDDEN|_A_SYSTEM|_A_SUBDIR|_A_ARCH;
    #define MAXPATH _MAX_PATH
  #endif    
#endif

#if defined(_MSC_VER) || defined(LINUX)
  const char* const* _argv;
  int _argc;
#endif

#if defined(LINUX)
  struct ffblk{
    char *ff_name;
    off_t ff_fsize;  
  };
#endif


struct FileInfo{
  char Name[MAXPATH+1];
};

// ******************************* memory allocation settings



#if defined(__MSDOS__)
  // System with 64k page limit
  const unsigned MaxFiles      =65532/4; // it uses *4 bytes for each file
  const int      MaxStartPaths =32;
  const size_t   CompFileSize  =1024*63,
                 FirstCompSize =1024;
#else
  // System with lots of memory and without 64 limit
  const unsigned long Megabyte =14360576;  
  const unsigned MaxFiles      =Megabyte/4;
  const int      MaxStartPaths =256;
  const size_t   CompFileSize  =Megabyte,
                 FirstCompSize =1024;
#endif
  
  
unsigned ClusterSize=4096;
int ReportEmpty=0;  

//************************************************ global vars
static unsigned long* FileLen; // file lenghts array
static unsigned Counter=0;     // number of files found
static int fnames=-1;          // temporary file handle
static char* fnames_n=NULL;    // temporary file name
static TLogOutput DupLog;      // report output stream
static char *buf1, *buf2;      // comparing buffers
BOOL CompareZeroFiles=FALSE;

static BOOL BreakSignated=FALSE; // "ESC pressed"
static BOOL PrintProgress=TRUE;
static BOOL PrintFiles=TRUE;
static BOOL PrintSummary=TRUE;


int OpenTemp(PChar& fn){
  #if !defined(LINUX)    
    char buf[MAXPATH+1];
    char* dir=getenv("TEMP");
    if(dir==NULL || !(*dir)) dir=getenv("TMP");
    if(dir==NULL || !(*dir)) dir=GetPath(_argv[0], buf);  
    fn=tempnam(dir, "~SGFS");
    return open(fn, O_RDWR|O_BINARY|O_CREAT|O_TRUNC);
  #else
    fn=new char[MAXPATH+1];
    strcpy(fn, "/tmp/~sg-findsim-XXXXXX");
    int ret=mkstemp(fn);
    //printf("Temp file: %s\n", fn);
    return ret;
  #endif    
}

//  0 - files are identical
//  1 - files are different
// -1 - error
int CompareFiles(const char* fn1, const char* fn2, unsigned long l){
  unsigned long BytesLeft=l;
  size_t readb=0;
  int f1, f2, retv=0;
  static int ClocksCnt=0;
  if((f1=open(fn1, O_RDONLY|O_BINARY))==-1){
    StdLog<<"Warning: unable to open "<<fn1<<"\n";
    return -1; };
  if((f2=open(fn2, O_RDONLY|O_BINARY))==-1){
    StdLog<<"Warning: unable to open "<<fn2<<"\n";
    close(f1);
    return -1; };
  while(BytesLeft){
    
    #if !defined(LINUX)
      if(kbhit()) if(getch()==ESC){ BreakSignated=TRUE; return 0; }
    #endif  
    
    if(!(ClocksCnt++&15) && PrintProgress) Clocks();
    if(!readb)
      readb=(ULONG(FirstCompSize)>BytesLeft)?BytesLeft:FirstCompSize;
    else
      readb=(BytesLeft>ULONG(CompFileSize))?CompFileSize:BytesLeft;
    if(size_t(read(f1, buf1, readb))!=readb){
      StdLog<<"Warning: error reading "<<fn1<<"\n";
      retv=-1;
      break; }
    if(size_t(read(f2, buf2, readb))!=readb){
      StdLog<<"Warning: error reading "<<fn2<<"\n";
      retv=-1;
      break; }
    if(memcmp(buf1, buf2, readb)){
      retv=1;
      break; }
    BytesLeft-=readb;
  };
  close(f1);
  close(f2);
  return retv;
}

void AddFile(const char* p, const struct ffblk* f){
  struct FileInfo fi;
  int sl;
//  if(f->ff_fsize || CompareZeroFiles){
  strcpy(fi.Name, p);
  sl=strlen(fi.Name);
  // Adding path separator to end of directory name
  if(!sl){ fi.Name[0]=PATH_CHAR; fi.Name[1]=0; }
  else if(sl>0) if(fi.Name[sl-1]!=PATH_CHAR){ fi.Name[sl]=PATH_CHAR; fi.Name[sl+1]=0; }      
  
  strcat(fi.Name, f->ff_name); 
  write(fnames, &fi, sizeof(FileInfo));
  if(Counter==MaxFiles) ExitProgram(2, "Fatal error: too many files!\n");
  FileLen[Counter++]=f->ff_fsize;
//  };
}

void RecurseScan(void){
  char Path[MAXPATH+1];
  static int ClocksCnt=0;
  if(!(ClocksCnt++&0x3) && PrintProgress) Clocks();
  #if !defined(LINUX)
    if(kbhit()) if(getch()==ESC){ BreakSignated=TRUE; return; }
  #endif    
  getcwd(Path, MAXPATH);
  #if defined(TRACE)
    printf("-- Scanning directory: %s\n", Path);
  #endif

  #if !defined(__MSDOS__)
    #if defined(LINUX)
      DIR *dir;
      dirent *de;
      struct ffblk ffb;
      struct stat st;      
      dir=opendir(".");
      #if defined(TRACE)
        printf("--- Opened, address: %ld\n", (long)dir);
      #endif              
      
      if(dir==NULL) return;
      while((de=readdir(dir))!=NULL){
        #if defined(TRACE)
          printf("---- Found: %s\n", de->d_name);
        #endif              	  
        if(strcmp(de->d_name, ".") && strcmp(de->d_name, "..")){
	  if(!chdir(de->d_name)){
            #if defined(TRACE)
              printf("---- ... is a directory\n");
            #endif              	  	  
            RecurseScan();
            if(BreakSignated) return;
            chdir(Path); 
	  }else{
            #if defined(TRACE)
              printf("---- ... is a file\n");
            #endif              	  	  	  
	    stat(de->d_name, &st);
            ffb.ff_name=de->d_name;
            ffb.ff_fsize=st.st_size;	    
            AddFile(Path, &ffb);	    
	  }    
        }
        #if defined(TRACE)
          printf("---- done with %s\n", de->d_name);
        #endif              	  	  	  	
      }					    
      closedir(dir);    
    #else
  
      struct ffblk ffb;  
      long handle;
      ffb.attrib=attr;
      if((handle=findfirst("*.*", &ffb))!=-1){
        if((ffb.attrib&_A_SUBDIR)==_A_SUBDIR){
          if(ffb.name[0]!='.')
            if(!chdir(ffb.name)){
              RecurseScan();
              if(BreakSignated) return;
              chdir(Path); }; }
        else AddFile(Path, &ffb);
        while(!findnext(handle, &ffb)){
          if((ffb.attrib&_A_SUBDIR)==_A_SUBDIR){
            if(ffb.name[0]!='.')
              if(!chdir(ffb.ff_name)){
                RecurseScan();
                if(BreakSignated) return;
                chdir(Path); }; }
          else AddFile(Path, &ffb);
        };
      };
    #endif  

  #else
    struct ffblk ffb;
    if(!findfirst("*.*", &ffb, attr)){
      if((ffb.ff_attrib&FA_DIREC)==FA_DIREC){
        if(ffb.ff_name[0]!='.')
          if(!chdir(ffb.ff_name)){
            RecurseScan();
            if(BreakSignated) return;
            chdir(Path); }; }
      else AddFile(Path, &ffb);
      while(!findnext(&ffb)){
        if((ffb.ff_attrib&FA_DIREC)==FA_DIREC){
          if(ffb.ff_name[0]!='.')
            if(!chdir(ffb.ff_name)){
              RecurseScan();
              if(BreakSignated) return;
              chdir(Path); }; }
        else AddFile(Path, &ffb);
      };
    };
  #endif
}

void FindSame(void){
  int i1, i2;
  int TotalSame=0, TotalWaste=0, ZeroFiles=0;
  long TotalCompared=0;
  unsigned long* p1, *p2;
  unsigned long DupSizeCount=0, WasteCount=0, ClusterWaste=0, csize;
  struct FileInfo f1, f2;
  for(i1=0, p1=FileLen; size_t(i1)<Counter; i1++, p1++) if(!(*p1)) ZeroFiles++;

  if(ReportEmpty && PrintFiles){
    if(ZeroFiles){
      if(PrintProgress) DupLog<<"\n*** FINDSIM Empty Files List: ***\n\n";
      for(i1=0, p1=FileLen; size_t(i1)<Counter; i1++, p1++) 
        if((*p1)==0){
          lseek(fnames, long(i1)*long(sizeof(FileInfo)), SEEK_SET);
          read(fnames, &f1, sizeof(FileInfo));
          DupLog<<f1.Name<<"\n";
        }
    }else{
      if(PrintProgress) DupLog<<"\n*** No empty files. ***\n\n";
    }  
  }

  if(PrintProgress) DupLog<<"\n*** Comparing files... ***\n";
  for(i1=0, p1=FileLen; size_t(i1)<Counter-1; i1++, p1++)
    if(*p1){
      csize=*p1;
      for(i2=i1+1, p2=p1+1; size_t(i2)<Counter; i2++, p2++)
        if(*p2 && csize==*p2){
          TotalCompared++;

          // read file names from temporary file
          lseek(fnames, long(i1)*long(sizeof(FileInfo)), SEEK_SET);
          read(fnames, &f1, sizeof(FileInfo));
          lseek(fnames, long(i2)*long(sizeof(FileInfo)), SEEK_SET);
          read(fnames, &f2, sizeof(FileInfo));
          if(!CompareFiles(f1.Name, f2.Name, csize)){
            if(BreakSignated) return;
            WasteCount+=csize;
            DupSizeCount+=csize;
            ClusterWaste+=(csize+ClusterSize-1)/ClusterSize;
            TotalSame++;
            TotalWaste++;
            if(*p1){
              if(PrintFiles) DupLog<<" \n"<<f1.Name<<"\n";
              *p1=0l;
              TotalSame++;
              DupSizeCount+=csize; }
            if(PrintFiles) DupLog<<f2.Name<<"\n";
            *p2=0l;
          }else if(BreakSignated) return;
        };
    }

  if(PrintSummary)
    DupLog<<" \n*** Summary: ***\n"
      "  Total files.........."<<long(Counter)<<"\n"
      "    Empty files........"<<int(ZeroFiles)<<"\n"
      "  Duplicated files....."<<int(TotalSame)<<"\n"
      "    Duplicated bytes..."<<long(DupSizeCount)<<"\n"
      "  Waste files.........."<<int(TotalWaste)<<"\n"
      "    Waste bytes........"<<long(WasteCount)<<"\n"
      "    Space wasted......."<<long(ClusterWaste*ClusterSize)<<
                               " bytes for cluster size=="<<
                               int(ClusterSize)<<"\n"
      "\n";
//    "    "<<long(TotalCompared)<<" are compared\n";
}

void ScanSubtree(const char *const start[]){
  int StartDisk;
  char RunPath[MAXPATH+1];
  // । 
  if(TryAlloc(3,
              PPChar(&FileLen), MaxFiles*sizeof(long),
              &buf1, CompFileSize,
              &buf2, CompFileSize, NULL)!=3)
    ExitProgram(ecUnrecOutOfMemory,
                "Not enough memory! I need about 256 kb.\n");
  //printf("= %lX = %lX = %lX =\n", FileLen, buf1, buf2);
  //getchar(); ExitProgram(0, "Ha!\n");}

  
  if((fnames=OpenTemp(fnames_n))==-1){
    StdLog<<"Unable to open temp file: "<<fnames_n<<"!\n";
    return; };
  
  // ***************************************************************** scanner
  

  #if !defined(LINUX)
    StdLog<<"Hit <ESC> to abort.\n";
  #endif    
  for(int i=0; start[i]!=NULL; i++){
    getcwd(RunPath, MAXPATH);
    #if !defined(LINUX)
      StartDisk=getdisk();
    #endif  
    if(PrintProgress) StdLog<<"Scanning path: "<<start[i]<<"\n";
    if(chdir(start[i])){
      if(PrintProgress) StdLog<<"Path not found: "<<start[i]<<"\n";
      return; };

    #if !defined(LINUX)
      if(start[i][1]==':') 
        #if defined(__MSDOS__)
          setdisk(toupper(start[i][0])-'A');
        #else
          setdisk(toupper(start[i][0])-'A'+1);
        #endif
    #endif	

    RecurseScan();
    chdir(RunPath);
    #if !defined(LINUX)
      setdisk(StartDisk);
    #endif  
    if(BreakSignated) break;
  }
  if(!BreakSignated){
    if(PrintProgress) StdLog<<"Total: "<<long(Counter)<<" file(s)\n";
    if(!Counter){
      if(PrintProgress) StdLog<<"No files to compare.\n";
    }else  
      FindSame();
  }

  close(fnames);
  unlink(fnames_n);
  #if defined(LINUX)
    delete fnames_n;
  #else
    free(fnames_n);
  #endif        

  delete FileLen, buf1, buf2;

  if(PrintProgress) StdLog<<((BreakSignated)?"Aborted by user!\n":"Done.\n");
}

int main(int argc, char* argv[]){
  char *Paths[MaxStartPaths];
  int PathCnt=0, i;

  #if defined(_MSC_VER) || defined(LINUX)
    _argc=argc;
    _argv=argv;
  #endif

  if(argc<2){
    StdLog<<"\n"
            "Findsim - Identical files finder " SIM_VERSION " "__DATE__" "__TIME__"\n"
            "(C) Sergey Galin, 1999-2004 http://sageshome.net Email: findsim@sageshome.net\n"
            "This program is distributed under GNU General Public License,\n"
	    "version 2.0 or higher, published by Free Software Foundation.\n"
	    "\n"
	    "Please read program documentation carefully before using this program.\n"
	    "There are some known issues.\n"
	    "\n"
            "USAGE:   findsim <startpath> [<startpath>...] [options]\n"
	    #if defined(LINUX)
              "EXAMPLE: findsim ./work -l=dupes.txt\n"
	    #else  
              "EXAMPLE: findsim c:\\work -l=dupes.txt\n"	    
	    #endif  
	      
            "\n"
	    "OPTIONS:\n"
            "-l=<dup_log> - duplicate log into file <dup_log>\n"
            "-e           - report empty (zero length) files\n"
	    
	    "-c=<bytes>   - set cluster size (for Summary)\n"
	    "-np          - do not print \"clock\" and progress messages to stdout\n"
	    "-nf          - do not print filenames\n"
	    "-ns          - do not print summary\n"
	    "\n"
	    "Note: options are case-sensitive.\n"
	    "\n"	    
	    ;

    return 1; };
  for(i=1; i<argc; i++)
    if(!memcmp(argv[i], "-l=", 3)){
      if(DupLog.AppendFile(argv[i]+3)){
         StdLog<<"Error opening file: "<<argv[i]+3<<"\n";
         return 2; }
    }else if(!memcmp(argv[i], "-c=", 3)){
      ClusterSize=atoi(argv[i]+3);
      if(ClusterSize<=0){ 
        StdLog<<"Bad cluster size specified.\n";
        return 1; }
    }else if(!strcmp(argv[i], "-e")){
      ReportEmpty=1;
    }else if(!strcmp(argv[i], "-np")){
      PrintProgress=FALSE;  
    }else if(!strcmp(argv[i], "-nf")){
      PrintFiles=FALSE;  
    }else if(!strcmp(argv[i], "-ns")){
      PrintSummary=FALSE;                 
    }else{
      if(PathCnt==MaxStartPaths-1){
        StdLog<<"Too many paths.\n";
        return 1; }
      else Paths[PathCnt++]=argv[i];
    }
    
  // Checking for bad parameters    
  if(!PathCnt){
    DupLog<<"No paths to scan specified.";
    DupLog.CloseFile();
    return 1;
  }    

  if(!PrintSummary && !PrintFiles){
    DupLog<<"Error: -nf and -ns options cannot be used simultaneously.\n";
    DupLog.CloseFile();
    return 1;  
  }
  
  Paths[PathCnt]=NULL;
  ScanSubtree(Paths); // scanner
  DupLog.CloseFile();
  return 0;
}

